from __future__ import annotations

from typing import Protocol, Any, Sequence
import math

import numpy as np
import scipy

from .utils import linear_assignment
from ..ops import iou_loss
from ..typing import ArrayLike


class KalmanFilter(Protocol):
    """KalmanFilter Protocol definition"""

    def initialize(self, measurement: Any)->tuple[np.ndarray, np.ndarray]:
        """Create track from unassociated measurement.

        Args:
            measurement : data in measurement space

        Returns:
            mean: Unobserved velocities are initialized to 0 mean.
            cov: covariance matrix (8x8 dimensional) of the new track. 

        """
        ...
    
    def predict(self, mean: Any, covariance: Any)->tuple[np.ndarray, np.ndarray]:
        """Run Kalman filter prediction step.

        Args:
            mean : mean vector of the object state at the previous time step.
            covariance : The covariance matrix of the object state at the
                previous time step.

        Returns:
            mean: the mean vector of the predicted state
            cov: the covariance matrix of the predicted
        """
        ...

    def update(self, mean: Any, covariance: Any, measurement: Any)->tuple[np.ndarray, np.ndarray]:
        """Run Kalman filter correction step.

        Args:
            mean :  The predicted state's mean vector.
            covariance : The state's covariance matrix.
            measurement : The measurement vector

        Returns:
            mean: mean vector of the measurement-corrected state distribution.
            cov: covariance matrix of the measurement-corrected state distribution.
        """
        ...


class ConstantVelocityKalmanFilter(KalmanFilter):
    """
    A simple Kalman filter for tracking bounding boxes in image space.  The 8-dimensional state space

        x, y, a, h, vx, vy, va, vh

    contains the bounding box center position (x, y), aspect ratio a, height h,
    and their respective velocities.

    Object motion follows a constant velocity model. The bounding box location
    (x, y, a, h) is taken as direct observation of the state space (linear
    observation model).
    """

    def __init__(self, *, std_weight_position:float = 1/20, std_weight_velocity:float =1/160):
        """
        Keyword Args:
            std_weight_position: Relative observation uncertainty.
            std_weight_velocity: Relative velocity uncertainty.
        """
        ndim, dt = 4, 1.

        # Create Kalman filter model matrices.
        self._motion_mat = np.eye(2 * ndim, 2 * ndim)
        for i in range(ndim):
            self._motion_mat[i, ndim + i] = dt
        self._update_mat = np.eye(ndim, 2 * ndim)

        self._std_weight_position = std_weight_position
        self._std_weight_velocity = std_weight_velocity

    def initiate(self, measurement:ArrayLike)->tuple[np.ndarray, np.ndarray]:
        mean_pos = measurement
        mean_vel = np.zeros_like(mean_pos)
        mean = np.r_[mean_pos, mean_vel]

        std = [
            2 * self._std_weight_position * measurement[3],
            2 * self._std_weight_position * measurement[3],
            2 * 1e-2,
            2 * self._std_weight_position * measurement[3],
            10 * self._std_weight_velocity * measurement[3],
            10 * self._std_weight_velocity * measurement[3],
            10 * 1e-5,
            10 * self._std_weight_velocity * measurement[3],
        ]
        covariance = np.diag(np.square(std))

        return mean, covariance

    def predict(self, mean:ArrayLike, covariance:ArrayLike)->tuple[np.ndarray, np.ndarray]:
        std_pos = [
            self._std_weight_position * mean[3],
            self._std_weight_position * mean[3],
            1e-2,
            self._std_weight_position * mean[3],
        ]
        std_vel = [
            self._std_weight_velocity * mean[3],
            self._std_weight_velocity * mean[3],
            1e-5,
            self._std_weight_velocity * mean[3],
        ]
        motion_cov = np.diag(np.square(np.r_[std_pos, std_vel]))

        mean = np.dot(mean, self._motion_mat.T)
        covariance = np.linalg.multi_dot((
            self._motion_mat, covariance, self._motion_mat.T)) + motion_cov

        return mean, covariance

    def project(self, mean:ArrayLike, covariance:ArrayLike) ->tuple[np.ndarray, np.ndarray]:
        std = [
            self._std_weight_position * mean[3],
            self._std_weight_position * mean[3],
            1e-1,
            self._std_weight_position * mean[3]]
        innovation_cov = np.diag(np.square(std))

        mean = np.dot(self._update_mat, mean)
        covariance = np.linalg.multi_dot((
            self._update_mat, covariance, self._update_mat.T))
        return mean, covariance + innovation_cov

    def multi_predict(self, mean:ArrayLike, covariance:ArrayLike)->tuple[np.ndarray, np.ndarray]:
        """vectorized version of the prediction step (Vectorized version).

        Args:
            mean : Nx8 dimensional mean matrix of the object states at the previous
                time step.
            covariance : Nx8x8 dimensional covariance matrics of the object states at the
                previous time step.

        Returns:
            mean: the mean vector of the predicted state
            cov:  the covariance matrix of the predicted state
        """
        std_pos = [
            self._std_weight_position * mean[:, 3],
            self._std_weight_position * mean[:, 3],
            1e-2 * np.ones_like(mean[:, 3]),
            self._std_weight_position * mean[:, 3]]
        std_vel = [
            self._std_weight_velocity * mean[:, 3],
            self._std_weight_velocity * mean[:, 3],
            1e-5 * np.ones_like(mean[:, 3]),
            self._std_weight_velocity * mean[:, 3]]
        sqr = np.square(np.r_[std_pos, std_vel]).T

        motion_cov = []
        for i in range(len(mean)):
            motion_cov.append(np.diag(sqr[i]))
        motion_cov = np.asarray(motion_cov)

        mean = np.dot(mean, self._motion_mat.T)
        left = np.dot(self._motion_mat, covariance).transpose((1, 0, 2))
        covariance = np.dot(left, self._motion_mat.T) + motion_cov

        return mean, covariance

    def update(self, mean:ArrayLike, covariance:ArrayLike, measurement:ArrayLike)->tuple[np.ndarray, np.ndarray]:
        projected_mean, projected_cov = self.project(mean, covariance)

        chol_factor, lower = scipy.linalg.cho_factor(
            projected_cov, lower=True, check_finite=False)
        kalman_gain = scipy.linalg.cho_solve(
            (chol_factor, lower), np.dot(covariance, self._update_mat.T).T,
            check_finite=False).T
        innovation = measurement - projected_mean

        new_mean = mean + np.dot(innovation, kalman_gain.T)
        new_covariance = covariance - np.linalg.multi_dot((
            kalman_gain, projected_cov, kalman_gain.T))

        return new_mean, new_covariance

    def gating_distance(self, mean, covariance, measurements,
                        only_position:bool|None=False, metric:str='maha'
        ) -> np.ndarray:
        """ Compute gating distance between state distribution and measurements.
            A suitable distance threshold can be obtained from `chi2inv95`. If
            `only_position` is False, the chi-square distribution has 4 degrees of
            freedom, otherwise 2.
        
        Args:
            mean: Mean vector over the state distribution (8 dimensional).
            covariance : Covariance of the state distribution (8x8 dimensional).
            measurements : An Nx4 dimensional matrix of N measurements, each in
                format (x, y, a, h) where (x, y) is the bounding box center
                position, a the aspect ratio, and h the height.
            only_position : If True, distance computation is done with respect to the bounding
                box center position only.

        Returns: 
            gating_distance: an array of length N, where the i-th element contains the
                squared Mahalanobis distance between (mean, covariance) and `measurements[i]`.
        """
        mean, covariance = self.project(mean, covariance)
        if only_position:
            mean, covariance = mean[:2], covariance[:2, :2]
            measurements = measurements[:, :2]

        d = measurements - mean
        if metric == 'gaussian':
            return np.sum(d * d, axis=1)
        elif metric == 'maha':
            cholesky_factor = np.linalg.cholesky(covariance)
            z = scipy.linalg.solve_triangular(
                cholesky_factor, d.T, lower=True, check_finite=False,
                overwrite_b=True)
            squared_maha = np.sum(z * z, axis=0)
            return squared_maha
        else:
            raise ValueError('invalid distance metric')


class ConstantVelocityKalmanFilter3D(KalmanFilter):
    """A 3D Kalman filter for tracking 3D bounding boxes follows a constant velocity model. 
        The 8-dimensional state space are:

            z, y, x, d, h, w, vz, vy, vx, vd, vh, vw
    """
    def __init__(self, *, std_weight_position:float=1/20, std_weight_velocity:float=1/160, z_scaling:float=1):
        """
        Keyword Args:
            std_weight_position: Relative observation uncertainty.
            std_weight_velocity: Relative velocity uncertainty.
            z_scaling: relative pixel size along z-axis vs x-y-axis
        """
        ndim, dt = 6, 1.

        # Create Kalman filter model matrices.
        self._motion_mat = np.eye(2 * ndim, 2 * ndim)
        for i in range(ndim):
            self._motion_mat[i, ndim + i] = dt
        self._update_mat = np.eye(ndim, 2 * ndim)

        self._std_weight_position = np.array([
            std_weight_position / z_scaling,
            std_weight_position,
            std_weight_position,
            std_weight_position / z_scaling,
            std_weight_position,
            std_weight_position,
        ])            

        self._std_weight_velocity = np.array([
            std_weight_velocity / z_scaling,
            std_weight_velocity,
            std_weight_velocity,
            std_weight_velocity / z_scaling,
            std_weight_velocity,
            std_weight_velocity
        ]) 

    def initiate(self, measurement:ArrayLike)->tuple[np.ndarray, np.ndarray]:
        mean_pos = measurement
        mean_vel = np.zeros_like(mean_pos)
        mean = np.r_[mean_pos, mean_vel]

        mean_size = (math.prod(mean[4:6])) ** (1/2)
        std = np.r_[
            2 * self._std_weight_position * mean_size,
            10 * self._std_weight_velocity * mean_size,
        ]
        covariance = np.diag(np.square(std))

        return mean, covariance

    def predict(self, mean:ArrayLike, covariance:ArrayLike)->tuple[np.ndarray, np.ndarray]:
        mean_size = (math.prod(mean[4:6])) ** (1/2)

        std_pos = self._std_weight_position * mean_size
        std_vel = self._std_weight_velocity * mean_size

        motion_cov = np.diag(np.square(np.r_[std_pos, std_vel]))

        mean = np.dot(mean, self._motion_mat.T)
        covariance = np.linalg.multi_dot((
            self._motion_mat, covariance, self._motion_mat.T)) + motion_cov

        return mean, covariance

    def project(self, mean:ArrayLike, covariance:ArrayLike)->tuple[np.ndarray, np.ndarray]:
        mean_size = (math.prod(mean[4:6])) ** (1/2)

        std = self._std_weight_position * mean_size

        innovation_cov = np.diag(np.square(std))

        mean = np.dot(self._update_mat, mean)
        covariance = np.linalg.multi_dot((
            self._update_mat, covariance, self._update_mat.T))
        return mean, covariance + innovation_cov

    def multi_predict(self, mean:ArrayLike, covariance:ArrayLike)->tuple[np.ndarray, np.ndarray]:
        """Vectorized Kalman filter prediction step.
        """
        mean_size = np.sqrt(mean[:, -1] * mean[:, -2])

        std_pos = self._std_weight_position[:, None] * mean_size
        std_vel = self._std_weight_velocity[:, None] * mean_size

        sqr = np.square(np.r_[std_pos, std_vel]).T

        motion_cov = []
        for i in range(len(mean)):
            motion_cov.append(np.diag(sqr[i]))
        motion_cov = np.asarray(motion_cov)

        mean = np.dot(mean, self._motion_mat.T)
        left = np.dot(self._motion_mat, covariance).transpose((1, 0, 2))
        covariance = np.dot(left, self._motion_mat.T) + motion_cov

        return mean, covariance

    def update(self, mean:ArrayLike, covariance:ArrayLike, measurement:ArrayLike)->tuple[np.ndarray, np.ndarray]:
        projected_mean, projected_cov = self.project(mean, covariance)

        chol_factor, lower = scipy.linalg.cho_factor(
            projected_cov, lower=True, check_finite=False)
        kalman_gain = scipy.linalg.cho_solve(
            (chol_factor, lower), np.dot(covariance, self._update_mat.T).T,
            check_finite=False).T
        innovation = measurement - projected_mean

        new_mean = mean + np.dot(innovation, kalman_gain.T)
        new_covariance = covariance - np.linalg.multi_dot((
            kalman_gain, projected_cov, kalman_gain.T))

        return new_mean, new_covariance

class KTracker:
    """ A stateful KalmanFilter tracker
    """
    _count = -1
 
    @classmethod
    def next_id(cls):
        cls._count += 1
        return cls._count

    @property
    def mean(self):
        try:
            return self._mean.copy()
        except AttributeError:
            return None

    @property
    def cov(self):
        try:
            return self._cov.copy()
        except AttributeError:
            return None
    

    def __repr__(self):
        if self.state == "uninitialized":
            return f"KT_U_{self.frame_id}"
        else:
            return f'KT_{self.track_id}_({self.start_frame}-{self.frame_id})'


    def __init__(self, init_obs:ArrayLike, frame_id:int, kf: KalmanFilter, *, data:dict={}):
        """ constructor

        Args:
            init_obs: initial measurement space data
            frame_id: the frame number
            kf: a Kalman Filter following the KalmanFilter protocol
        
        Keyword Args:
            data: additional data of the observation that will be store
        """
        self.obs = np.asarray(init_obs, dtype=float)
        self._data = data
        self.tracklet_len = 0
        self.kf = kf
        self.frame_id = frame_id
        self.start_frame = frame_id
        self.state = "uninitialized"


    def predict(self):
        """ Kalman filter prediction
        """
        self._mean, self._cov = self.kf.predict(
            self.mean, self.cov,
        )


    def initialize(self):
        """ Initialze the track, so that it has a track id and a history object
        """
        self._mean, self._cov = self.kf.initiate(
            self.obs,
        )            
        self.track_id = self.next_id()
        self._data.update(dict(
            frame_id = self.frame_id,
            predicted_obs = self.mean,
        ))
        self.history = [self._data]
        self.state = "tracked"


    def update(self, new_track: KTracker):
        """ Kalman filter update step

        new_track: the new observation
        """
        self._mean, self._covariance = self.kf.update(
            self.mean, self.cov, new_track.obs
        )

        self.frame_id = new_track.frame_id
        self.obs = new_track.obs

        if self.state == "tracked":
            self.tracklet_len += 1
        else:
            self.tracklet_len = 0
        
        self.state = "tracked"
        data = new_track._data
        data.update(dict(
            frame_id = new_track.frame_id,
            preicted_obs = self.mean,
        )) 
        self.history.append(new_track._data)


    def new_id(self):
        self.track_id = self.next_id() 


    def mark_lost(self, *, update_state:ArrayLike=None):
        """ Mark the track as lost

        Keyword Args:
            update_state: if not None, change the mean vector to this one
        """
        self.state = "lost"
        if update_state is not None:
            self._mean = np.array(update_state)


    def mark_removed(self):
        """ Mark the track as removed
        """
        self.state = "removed"

    @staticmethod
    def assign(
        tracks: Sequence[KTracker], 
        dets: Sequence[KTracker], 
        cost_matrix: ArrayLike,
        threshold:float,
    ) -> tuple[Sequence[KTracker], Sequence[KTracker], Sequence[KTracker]]:
        """ perform MOT assigment

        Args:
            tracks: current list of tracks of length M
            dets: list of new observations of length N
            cost_matrix: [M x N] cost matrix
            threshold: max cost for linking
        
        Returns:
            tracked: list of tracks that are updated with new observations
            remaining_tracks: list of tracks that did not find a suitable link
            remaining_dets: the remaining detections that didn't get linked in.            
        """
        matches, unmatched_track_ids, unmatched_det_ids = linear_assignment(cost_matrix, thresh=threshold)

        tracked = []
        for itracked, idet in matches:
            track = tracks[itracked]
            track.update(dets[idet])
            tracked.append(track)            

        remaining_tracks = [tracks[it] for it in unmatched_track_ids]
        remaining_dets = [dets[it] for it in unmatched_det_ids]

        return tracked, remaining_tracks, remaining_dets



import logging
import threading
import traceback
from concurrent import futures
from pathlib import Path

from typing import Iterable

import grpc
import jax
import numpy as np
import typer

from . import proto
from .common import get_dtype, decode_image, TokenValidationInterceptor, LacssServicerBase

app = typer.Typer(pretty_exceptions_enable=False)

_MAX_MSG_SIZE=1024*1024*128
_TARGET_CELL_SIZE=32
_MAX_STREAM_MSG_SIZE = _MAX_MSG_SIZE * 16
MAX_IMG_SIZE = 1024 * 1024 * 1024

def _process_input(request: proto.DetectionRequest, image=None):
    pixels = request.image_data.pixels
    settings = request.detection_settings

    if image is None:
        image = decode_image(pixels)

    physical_size = np.array([
        pixels.physical_size_z or pixels.physical_size_x, # one might set xy but not z
        pixels.physical_size_y, 
        pixels.physical_size_x,
    ], dtype="float")
    if (physical_size == 0).any():
        physical_size[:] = 1.0

    if settings.HasField("cell_diameter_hint"):
        scaling = _TARGET_CELL_SIZE / settings.cell_diameter_hint * physical_size

    else:
        if physical_size[1] != physical_size[2]:
            raise ValueError("Scaling hint provided, but pixel is not isometric")

        scaling = np.array([ settings.scaling_hint or 1.0 ] * 3, dtype="float")
        scaling[0] *= physical_size[0] / physical_size[1] 
    
    logging.info(f"Requested rescaling factor is {scaling}")

    shape_hint = tuple( np.round(scaling * image.shape[:3]).astype(int) )

    if image.shape[0] == 1: # 2D
        image = image.squeeze(0)
        shape_hint = shape_hint[1:]

    kwargs = dict(
        reshape_to = shape_hint,
        score_threshold = settings.min_score or 0.4,
        min_area = settings.min_cell_area,
        nms_iou = settings.nms_iou,
        segmentation_threshold = settings.segmentation_threshold or 0.5,
    )

    return image, kwargs


def _process_result(preds, image) -> proto.DetectionResponse:
    response = proto.DetectionResponse()

    if image.ndim == 3: # returns polygon

        for contour, score in zip(preds["pred_contours"], preds["pred_scores"]):
            if len(contour) == 0:
                continue

            scored_roi = proto.ScoredROI(
                score = score,
                roi = proto.ROI(
                    polygon = proto.Polygon(points = [proto.Point(x=p[0], y=p[1]) for p in contour]),
                )
            )

            response.detections.append(scored_roi)

    else: # 3d returns Mesh
        for mesh, score in zip(preds["pred_contours"], preds["pred_scores"]):
            scored_roi = proto.ScoredROI(
                score = score,
                roi = proto.ROI(
                    mesh = proto.Mesh(
                        verts = [proto.Point(z=v[0], y=v[1], x=v[2]) for v in mesh['verts']],
                        faces = [proto.Mesh.Face(p1=p[0], p2=p[1], p3=p[2]) for p in mesh['faces']],
                    ),
                )
            )

            response.detections.append(scored_roi)

    return response


def _process_grid_input(request_iterator:Iterable[proto.DetectionRequest]):
    d, h, w, c = 0, 0, 0, 0
    images, grids = [], []
    request = None
    total_msg_size = 0
    for request in request_iterator:
        total_msg_size += request.ByteSize()
        if total_msg_size > _MAX_STREAM_MSG_SIZE:
            raise ValueError("input message size  {total_msg_size} exceeded limit.")

        pixels = request.image_data.pixels

        image = decode_image(pixels)

        grids.append([
            pixels.offset_z,
            pixels.offset_y,
            pixels.offset_x,
        ])
        images.append(image)

        d = max(d, pixels.offset_z + image.shape[0])
        h = max(d, pixels.offset_z + image.shape[1])
        w = max(d, pixels.offset_z + image.shape[2])
        c = max(c, image.shape[3])

        if d * h * w > MAX_IMG_SIZE:
            raise ValueError(f"input image is too large {(d, h, w)}")

    # empty input iterator
    if request is None:
        return None, None

    assert c <= 3

    full_image = np.zeros([d, h, w, c], dtype=images[0].dtype)
    for image, grid in zip(images, grids):
        full_image[
            grid[0]:grid[0]+image.shape[0],
            grid[1]:grid[1]+image.shape[1],
            grid[2]:grid[2]+image.shape[2],
            :image.shape[3],
        ] = image

    return _process_input(request, full_image)


class LacssServicer(LacssServicerBase):

    def __init__(self, model):
        super().__init__()

        self.model = model

    def RunDetection(self, request, context):
        logging.info(f"Received message of size {request.ByteSize()}")

        try:
            image, kwargs = _process_input(request)

            logging.info(f"received image {image.shape}")

        except Exception as e:
            logging.error(repr(e))

            logging.error(traceback.format_exc())

            context.abort(grpc.StatusCode.INVALID_ARGUMENT, repr(e))

            return

        try:    
            with self._lock:
                preds = self.model.predict(
                    image, output_type="contour", **kwargs,
                )

            response = _process_result(preds, image)

            logging.info(f"Reply with message of size {response.ByteSize()}")

            return response
        
        except Exception as e:

            logging.error(repr(e))

            logging.error(traceback.format_exc())

            context.abort(grpc.StatusCode.UNKNOWN, f"prediction failed with error: {repr(e)}")



    def RunDetectionOnGrid(self, request_iterator, context):
        try:
            image, kwargs = _process_grid_input(request_iterator)

            if image is None:
                return proto.DetectionResponse()

            logging.info(f"Received full image of size {image.shape[:-1]}")

            with self._lock:
                preds = self.model.predict(
                    image, output_type="contour", **kwargs,
                )

            response = _process_result(preds, image)

            logging.info(f"Reply with message of size {response.ByteSize()}")

            return response

        except ValueError as e:
            
            logging.error(repr(e))

            logging.error(traceback.format_exc())

            context.abort(grpc.StatusCode.INVALID_ARGUMENT, repr(e))

        except Exception as e:

            logging.error(repr(e))

            logging.error(traceback.format_exc())

            context.abort(grpc.StatusCode.UNKNOWN, f"prediction failed with error: {repr(e)}")


def get_predictor(modelpath, f16):
    from .predict import Predictor

    model = Predictor(modelpath, f16=f16)

    logging.info(f"lacss_server: loaded model from {modelpath}")

    model.module.detector.max_output = 512  # FIXME good default?
    model.module.detector.min_score = 0.2

    logging.debug(f"lacss_server: precompile for most common shapes")

    _ = model.predict(np.ones([256,256,256,3]), output_type="contour")
    _ = model.predict(np.ones([384,384,384,3]), output_type="contour")
    _ = model.predict(np.ones([512,512,3]), output_type="contour")
    _ = model.predict(np.ones([1024,1024,3]), output_type="contour")

    return model


def show_urls():
    from . import model_urls
    
    print("Pretrained model files:")
    print("==============================")
    for k, v in model_urls.items():
        print(f"{k}: {v}")
    print()


@app.command()
def main(
    modelpath: Path|None = None,
    port: int = 50051,
    workers: int = 10,
    ip: str = "0.0.0.0",
    local: bool = False,
    token: bool|None = None,
    debug: bool = False,
    compression: bool = True,
    f16: bool = False,
):
    logging.basicConfig(level=logging.DEBUG if debug else logging.INFO)

    if modelpath is None:
        show_urls()
        return

    print ("server starting ...")

    model = get_predictor(modelpath, f16)

    logging.info(f"lacss_server: default backend is {jax.default_backend()}")

    if jax.default_backend() == "cpu":
        logging.warning(
            f"lacss_server: WARNING: No GPU configuration. This might be very slow ..."
        )

    if token is None:
        token = not local
    if token:
        import secrets

        token_str = secrets.token_urlsafe(64)

        print()
        print("COPY THE TOKEN BELOW FOR ACCESS.")
        print("=======================================================================")
        print(f"{token_str}")
        print("=======================================================================")
        print()
    else:
        token_str = None

    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=workers),
        compression=grpc.Compression.Gzip if compression else grpc.Compression.NoCompression,
        interceptors=(TokenValidationInterceptor(token_str),),
        options=(("grpc.max_receive_message_length", _MAX_MSG_SIZE),),
    )

    proto.add_LacssServicer_to_server(
        LacssServicer(model), 
        server,
    )

    if local:
        server.add_secure_port(f"127.0.0.1:{port}", grpc.local_server_credentials())
    else:
        server.add_insecure_port(f"{ip}:{port}")

    logging.info(f"lacss_server: listening on port {port}")

    print ("server starting ... ready")

    server.start()
    server.wait_for_termination()


if __name__ == "__main__":
    app()



from __future__ import annotations

import tensorflow as tf
from .generator import _crop_and_resize

def _init_boolean_mask(inputs):
    if "centroids" in inputs:
        return tf.ones([len(inputs["centroids"])], dtype=tf.bool)
    elif "bboxes" in inputs:
        return tf.ones([len(inputs["bboxes"])], dtype=tf.bool)
    else:
        return None


def _value_pair(x):
    try:
        a, b = x
    except:
        a = x
        b = x

    return a, b


def _image_size(img):
    H = tf.shape(img)[-3]
    W = tf.shape(img)[-2]

    return (H, W)

def _resize_3d(img, target_size, **kwargs):
    assert img.ndim == 4
    assert len(target_size) == 3
    img = tf.image.resize(img, target_size[1:], **kwargs)
    img = tf.transpose(img, (2,0,1,3))
    img = tf.image.resize(img, target_size[:2], **kwargs)
    img = tf.transpose(img, (1,2,0,3))
    return img

def resize_img(img, target_size, **kwargs):
    if img.ndim == 3:
        return tf.image.resize(img, target_size, **kwargs)
    else:
        return _resize_3d(img, target_size, **kwargs)


def resize(inputs: dict, *, target_size: tuple[int, int], p: float = 1.0) -> dict:
    """Resize image and labels

    Args:
        inputs: data from TF dataset. Affected elements are:

            * image
            * image_mask
            * centroids
            * bboxes

    Keyword Args:
        target_size: target size
        p: probability of applying transformation
    """

    def _resize(inputs):
        img = inputs['image']
        img_shape = tf.convert_to_tensor(tf.shape(img[..., 0]))
        target_sz = tf.convert_to_tensor(target_size)
        scalings = tf.cast(target_sz, tf.float32) / tf.cast(img_shape, tf.float32)

        inputs['image'] = resize_img(img, target_size, antialias=True)

        if "image_mask" in inputs:
            image_mask = inputs["image_mask"][..., None]  # need a channel dimension
            image_mask = resize_img(image_mask, target_size)
            inputs["image_mask"] = tf.squeeze(image_mask, axis=-1)

        if "centroids" in inputs:
            inputs["centroids"] = tf.cast(inputs["centroids"], dtype=tf.float32) * scalings

        if "bboxes" in inputs:
            inputs["bboxes"] = tf.cast(inputs["bboxes"], dtype=tf.float32) * tf.concat([scalings, scalings], axis=0)

        return inputs

    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs

    else:
        return _resize(inputs)


def random_resize(
    inputs: dict,
    *,
    scaling: float | tuple[float, float],
    keep_aspect_ratio: bool = False,
    p: float = 1.0,
) -> dict:
    """Resize image by a random amount

    Args:
        inputs: data from TF dataset. Affected elements are:

            * image
            * image_mask
            * centroids

    Keyword Args:
        scaling: range of scale, e.g, [0.8, 1.5].
        keep_aspect_ratio: Whether to scale x/y the same amount.
        p: probability of applying transformation

    """

    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs

    else:
        min_scale, max_scale = scaling

        img_shape = tf.convert_to_tensor(tf.shape(inputs['image'][..., 0]))
        dim = inputs['image'].ndim - 1

        H, W = _image_size(inputs["image"])

        if keep_aspect_ratio:
            scalings = tf.random.uniform([1], min_scale, max_scale)
            # scaling_y = tf.random.uniform([], min_scale, max_scale)
            # scaling_x = scaling_y
        else:
            scalings = tf.random.uniform([dim], min_scale, max_scale)

        target_size = tf.cast(tf.round(img_shape * scaling), tf.int32)

        return resize(inputs, target_size=target_size)


def crop_to_roi(
    inputs: dict,
    *,
    roi: tuple[int, ...],
    area_ratio_threshold: float = 1.0,
    clip_boxes: bool = True,
    p: float = 1.0,
) -> dict:
    """Crop image to bounding-box ROI

    Args:
        inputs: data from TF dataset. Affected elements are:

            * image
            * image_mask
            * centroids
            * bboxes
            * masks

    Keyword Args:
        roi: Rectangle roi in yxyx format
        area_ratio_threshold: remove instances if the bbox's relative remaining area is below this threshold
        p: probability of applying transformation

    """

    def _crop_to_roi(inputs):
        dim = inputs['image'].ndim - 1
        roi_ = tf.convert_to_tensor(roi)
        p0 = tf.cast(roi_[:dim], tf.float32)
        p1 = tf.cast(roi_[dim:], tf.float32)
        span = p1 - p0

        slices = [slice(int(p0[k]), int(p1[k])) for k in range(dim)]
        inputs['image'] = inputs['image'].__getitem__(slices)

        if "image_mask" in inputs:
            inputs["image_mask"] = inputs["image_mask"].__getitem__(slices)
        
        boolean_mask = _init_boolean_mask(inputs)
        if boolean_mask is None:
            return inputs

        if "centroids" in inputs:
            # ctrds = tf.cast(inputs["centroids"], tf.float32) - tf.cast(p0, tf.float32)
            ctrds = tf.cast(inputs["centroids"], tf.float32)
            boolean_mask &= tf.reduce_all(ctrds > p0, axis=-1) 
            boolean_mask &= tf.reduce_all(ctrds < p1, axis=-1)

        if "bboxes" in inputs:
            bboxes = tf.cast(inputs["bboxes"], tf.float32)
            bboxes = bboxes - tf.concat([p0, p0], axis=0)
            areas = tf.reduce_prod(bboxes[:, dim:] - bboxes[:, :dim], axis=-1)

            clipped_bboxes = tf.minimum(bboxes, tf.concat([span, span], axis=0))
            clipped_bboxes = tf.maximum(clipped_bboxes, 0)
            clipped_areas = tf.reduce_prod(clipped_bboxes[:, dim:] - clipped_bboxes[:, :dim], axis=-1)

            boolean_mask &= (clipped_areas / areas) >= area_ratio_threshold

        # so that TF won't complain boolean_mask has no shape
        boolean_mask = tf.ensure_shape(boolean_mask, [None])

        if "centroids" in inputs:
            inputs["centroids"] = tf.boolean_mask(ctrds, boolean_mask) - p0

        if "bboxes" in inputs:
            if clip_boxes:
                bboxes = tf.boolean_mask(bboxes, boolean_mask)
                inputs["bboxes"] = tf.boolean_mask(clipped_bboxes, boolean_mask)
            else:
                inputs["bboxes"] = tf.boolean_mask(bboxes, boolean_mask)

            if "masks" in inputs:
                if clip_boxes:
                    mask_shape = inputs["masks"][0].shape

                    bboxe_p0s = bboxes[:, :dim]
                    bboxe_szs = bboxes[:, dim:] - bboxes[:, :dim]

                    new_bbox = inputs["bboxes"] - tf.concat([bboxe_p0s, bboxe_p0s], axis=-1)
                    new_bbox = new_bbox / tf.concat([bboxe_szs, bboxe_szs], axis=-1)
                    new_bbox = new_bbox * tf.cast(
                        mask_shape + mask_shape, new_bbox.dtype
                    )
                    inputs["masks"] = _crop_and_resize(
                        tf.boolean_mask(inputs["masks"], boolean_mask),
                        new_bbox, mask_shape
                    )
                else:
                    mask_shape = inputs['masks'].shape
                    inputs["masks"] = tf.ensure_shape(
                        tf.boolean_mask(inputs["masks"], boolean_mask),
                        mask_shape,
                    )

        return inputs

    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs

    else:
        return _crop_to_roi(inputs)


def random_crop(
    inputs: dict,
    *,
    target_size: int|tuple[int, ...],
    area_ratio_threshold: float = 1.0,
    clip_boxes: bool = True,
    p: float = 1.0,
) -> dict:
    """Random crop to a set target size

    Args:
        inputs: data from TF dataset. Affected elements are:

            * image
            * image_mask
            * centroids
            * bboxes
            * masks

    Keyword Args:
        target_size: the target size
        area_ratio_threshold: remove instances if the bbox's relative remaining area is below this threshold
        p: probability of applying transformation
    """

    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs

    else:
        dim = inputs['image'].ndim - 1
        img_shape = tf.shape(inputs['image'])[:-1]
        target_size = tf.broadcast_to(
            tf.convert_to_tensor(target_size, dtype=tf.int32),
            [dim],
        )

        p0 = tf.random.uniform([dim]) * tf.cast(img_shape - target_size, tf.float32)
        p0 = tf.cast(p0, tf.int32)
        roi = tf.concat([p0, p0 + target_size], axis=0)
        return crop_to_roi(
            inputs, 
            roi=roi,
            area_ratio_threshold=area_ratio_threshold,
            clip_boxes=clip_boxes,
        )


def pad(
    inputs: dict,
    *,
    paddings: int | tuple[int, ...],
    constant_values: float = 0,
    p: float = 1.0,
) -> dict:
    """Pad image and labels.

    Args:
        inputs: data from TF dataset. Affected elements are:

            * image
            * image_mask
            * centroids
            * bboxes

    Keyword Args:
        paddings: either a tuple or a single value. If latter, use the same padding for both x and y axis
        constant_values: the value to fill the padded area
        p: probability of applying transformation
    """

    def _pad(inputs):
        dim = inputs['image'].ndim - 1
        paddings_ = tf.convert_to_tensor(paddings, dtype=tf.int32)
        paddings_ = tf.broadcast_to(paddings_, (dim, 2))

        padding_values = tf.concat([paddings_, tf.zeros([1,2], dtype=tf.int32)], axis=0)
        inputs["image"] = tf.pad(inputs["image"], padding_values, constant_values=constant_values,)

        if "image_mask" in inputs:
            image_mask = inputs["image_mask"]
            if image_mask.ndim == dim:
                padding_values = paddings_
            else:
                padding_values = tf.concat(
                    [tf.zeros([1,2], dtype=tf.int32), paddings_],
                    axis=0,
                )
            inputs["image_mask"] = tf.pad(
                inputs["image_mask"],
                padding_values,
                constant_values=constant_values,
            )

        if "centroids" in inputs:
            inputs["centroids"] += tf.cast(paddings_[:, 0], tf.float32)

        if "bboxes" in inputs:
            inputs["bboxes"] += tf.cast(
                tf.concat([paddings_[:, 0], paddings_[:, 0]], axis=0), tf.float32
            )

        return inputs

    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs

    else:
        return _pad(inputs)


def pad_to_size(
    inputs: dict,
    *,
    target_size: int|tuple[int, ...],
    constant_values: float = 0,
    p: float = 1.0,
) -> dict:
    """Pad image and labels to a target size. Padding is applied so that the orginal scene is centered in the output.

    Args:
        inputs: data from TF dataset. Affected elements are:

            * image
            * image_mask
            * centroids
            * bboxes

    Keyword Args:
        target_size: target image size
        constant_values: the value to fill the padded area
        p: probability of applying transformation
    """

    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs

    dim = inputs['image'].ndim -1
    img_shape = tf.shape(inputs['image'][..., 0])
    target_size = tf.broadcast_to(
        tf.convert_to_tensor(target_size, dtype=tf.int32),
        [dim],
    )

    paddings = tf.stack([
        tf.zeros([dim], dtype=tf.int32),
        target_size - img_shape    
    ], axis=-1)

    return pad(inputs, paddings=paddings, constant_values=constant_values)


def random_crop_or_pad(
    inputs: dict,
    *,
    target_size: tuple[int, int],
    constant_values: float = 0,
    area_ratio_threshold: float = 1.0,
    clip_boxes:bool = True,
    p: float = 1.0,
) -> dict:
    """Random crop or pad image to specified target_size.

    Args:
        inputs: data from TF dataset. Affected elements are:

            * image
            * image_mask
            * centroids
            * bboxes
            * masks

    Keyword Args:
        target_size: target size
        constant_values: the value to fill the padded area
        area_ratio_threshold: remove instances if the bbox's relative remaining area is below this threshold
        p: probability of applying transformation
    """

    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs

    else:
        dim = inputs['image'].ndim -1
        img_shape = tf.shape(inputs['image'][..., 0])
        target_size = tf.broadcast_to(
            tf.convert_to_tensor(target_size, dtype=tf.int32),
            [dim],
        )
        
        padsize = tf.maximum(target_size, img_shape)        

        inputs = pad_to_size(
            inputs, target_size=padsize, constant_values=constant_values
        )

        inputs = random_crop(
            inputs, target_size=target_size, area_ratio_threshold=area_ratio_threshold, clip_boxes=clip_boxes,
        )

        return inputs


def flip_left_right(inputs: dict, *, p: float = 1.0) -> dict:
    """Flip image left-right

    Args:
        inputs: data from TF dataset. Affected elements are:

            * image
            * image_mask
            * centroids
            * bboxes

    Keyword Args:
        p: probability of applying transformation

    """

    def _flip_left_right(inputs):
        dim = inputs['image'].ndim -1
        W = tf.shape(inputs['image'])[-2]

        inputs["image"] = tf.image.flip_left_right(inputs["image"])

        if "image_mask" in inputs:
            image_mask = inputs["image_mask"][..., None]
            inputs["image_mask"] = tf.squeeze(
                tf.image.flip_left_right(image_mask),
                axis=-1,
            )

        sgn = tf.cast([1, -1] if dim==2 else [1, 1, -1], tf.float32)
        ofs = tf.cast([0, W] if dim==2 else [0, 0, W], tf.float32)
        if "centroids" in inputs:
            inputs["centroids"] = inputs["centroids"] * sgn + ofs

        if "bboxes" in inputs:
            inputs["bboxes"] = inputs["bboxes"] * tf.concat([sgn, sgn], axis=0) + tf.concat([ofs, ofs], axis=0)

        return inputs

    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs

    else:
        return _flip_left_right(inputs)


def flip_up_down(inputs: dict, *, p: float = 1.0) -> dict:
    """Flip image up-down

    Args:
        inputs: data from TF dataset. Affected elements are:

            * image
            * image_mask
            * centroids
            * bboxes

    Keyword Args:
        p: probability of applying transformation

    """

    def _flip_up_down(inputs):
        dim = inputs['image'].ndim -1
        H = tf.shape(inputs['image'])[-3]

        inputs["image"] = tf.image.flip_up_down(inputs["image"])

        if "image_mask" in inputs:
            image_mask = inputs["image_mask"][..., None]
            inputs["image_mask"] = tf.squeeze(
                tf.image.flip_up_down(image_mask),
                axis=-1,
            )

        sgn = tf.cast([-1, 1] if dim==2 else [1, -1, 1], tf.float32)
        ofs = tf.cast([H, 0] if dim==2 else [0, H, 0], tf.float32)
        if "centroids" in inputs:
            inputs["centroids"] = inputs["centroids"] * sgn + ofs

        if "bboxes" in inputs:
            inputs["bboxes"] = inputs["bboxes"] * tf.concat([sgn, sgn], axis=0) + tf.concat([ofs, ofs], axis=0)

        return inputs

    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs

    else:
        return _flip_up_down(inputs)

def flip_top_bottom(inputs: dict, *, p: float = 1.0) -> dict:
    """Flip image up-down

    Args:
        inputs: data from TF dataset. Affected elements are:

            * image
            * image_mask
            * centroids
            * bboxes

    Keyword Args:
        p: probability of applying transformation

    """
    def _flip_top_bottom(inputs):
        dim = inputs['image'].ndim -1
        assert dim == 3

        D = tf.shape(inputs['image'])[-4]

        inputs["image"] = tf.image.flip_up_down(inputs["image"])

        if "image_mask" in inputs:
            image_mask = inputs["image_mask"][..., None]
            inputs["image_mask"] = tf.squeeze(
                tf.image.flip_up_down(image_mask),
                axis=-1,
            )

        sgn = tf.cast([-1, 1, 1], tf.float32)
        ofs = tf.cast([D, 0, 0], tf.float32)
        if "centroids" in inputs:
            inputs["centroids"] = inputs["centroids"] * sgn + ofs

        if "bboxes" in inputs:
            inputs["bboxes"] = inputs["bboxes"] * tf.concat([sgn, sgn], axis=0) + tf.concat([ofs, ofs], axis=0)

        return inputs

    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs

    else:
        return _flip_top_bottom(inputs)


def transpose(inputs, *, axes, p=1):
    def _tranpose(inputs):
        axes_ = tf.convert_to_tensor(axes, dtype=tf.int32)
        dim = inputs['image'].ndim -1

        inputs['image'] = tf.transpose(
            inputs['image'], 
            tf.concat([axes_, [dim]], axis=0),
        )
        if "image_mask" in inputs:
            inputs['image_mask'] = tf.transpose(inputs['image_mask'], axes_)

        if "masks" in inputs:
            inputs['masks'] = tf.transpose(
                inputs['masks'], 
                tf.concat([[0], axes_ + 1], axis=0),
            )

        if "centroids" in inputs:
            inputs["centroids"] = tf.gather(inputs["centroids"], axes_, axis=1)

        if "bboxes" in inputs:
            inputs["bboxes"] = tf.gather(
                inputs["bboxes"], 
                tf.concat([axes_, axes_+dim], axis=0), 
                axis=1,
            )

        return inputs

    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs
    else:
        return _tranpose(inputs)
    

def mosaic(inputs):
    img = inputs['image']
    H = tf.shape(img)[-3]
    W = tf.shape(img)[-2]
    img = tf.concat([
        tf.concat([img[0], img[1]], axis=-2),
        tf.concat([img[2], img[3]], axis=-2),
    ], axis=-3)
    inputs['image']=img
    
    ofs = tf.cast([
        [0,0],[0,W],[H, 0],[H, W]
    ], dtype=inputs['centroids'].dtype)
    ofs = tf.reshape(ofs, [4, 1, 2])
    locs = inputs['centroids'] + ofs
    inputs['centroids'] = locs.merge_dims(0,1)
    
    if "bboxes" in inputs:
        ofs_box = tf.concat([ofs, ofs], axis=-1)
        box = inputs['bboxes'] + ofs_box
        inputs['bboxes'] = box.merge_dims(0,1)
    
    if "masks" in inputs:
        inputs['masks'] = inputs['masks'].merge_dims(0,1)
    
    return inputs
    

@tf.py_function(Tout=tf.float32)
def _py_cutout(image, bbox, size, n):
    import numpy as np
    image = image.numpy()
    bbox = bbox.numpy()
    size = int(size)
    n = int(n)
    img_shape = image.shape[:-1]
    size = np.array(size).astype(int)

    sel = np.zeros([bbox.shape[0]], dtype=int)
    sel[:n] = 1
    np.random.shuffle(sel)
    sel_boxes = bbox[sel > 0]
    dim = sel_boxes.shape[-1]//2
    mins = np.minimum(sel_boxes[:,:dim], sel_boxes[:,dim:])
    maxs = np.maximum(sel_boxes[:,:dim], sel_boxes[:,dim:])
    c0s = np.random.randint(mins-size, maxs+size)

    for c0 in c0s:
        s = tuple(slice(max(0, a), min(c, b)) for a, b, c in zip(c0, c0+size, img_shape))
        image.__setitem__(s, 0)

    return image


def cutout(inputs, *, size:int=10, n:int=50, p:float=1.0):
    def _cutout(inputs):
        assert "bboxes" in inputs
        image = _py_cutout(inputs['image'], inputs['bboxes'], size, n)
        inputs['image'] = image
        return inputs
        
    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs
    else:
        return _cutout(inputs)
